/****************************************************************************

 This file is part of the GLC-lib library.
 Copyright (C) 2005-2008 Laurent Ribon (laumaya@users.sourceforge.net)
 http://glc-lib.sourceforge.net

 GLC-lib is free software; you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation; either version 3 of the License, or
 (at your option) any later version.

 GLC-lib is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with GLC-lib; if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

*****************************************************************************/

//! \file glc_world.cpp implementation of the GLC_World class.

#include "glc_world.h"
#include "glc_structinstance.h"
#include "glc_structreference.h"

GLC_World::GLC_World()
: m_pWorldHandle(new GLC_WorldHandle())
, m_pRoot(new GLC_StructOccurence())
{
	m_pRoot->setWorldHandle(m_pWorldHandle);
}

GLC_World::GLC_World(GLC_StructOccurence* pOcc)
: m_pWorldHandle(new GLC_WorldHandle())
, m_pRoot(pOcc)
{
	m_pRoot->setWorldHandle(m_pWorldHandle);
}

GLC_World::GLC_World(const GLC_World& world)
: m_pWorldHandle(world.m_pWorldHandle)
, m_pRoot(world.m_pRoot)
{
	//qDebug() << "GLC_World::GLC_World() : " << (*m_pNumberOfWorld) << " " << this;
	// Increment the number of world
	m_pWorldHandle->increment();

}

GLC_World::~GLC_World()
{
	// Decrement the number of world
	m_pWorldHandle->decrement();
	if (m_pWorldHandle->isOrphan())
	{
		delete m_pRoot;
		delete m_pWorldHandle;
    }
}

QList<GLC_StructOccurence *> GLC_World::minimumSelectedOccurenceList() const
{
    QSet<GLC_StructOccurence *> selectedOccSet= QSet<GLC_StructOccurence *>::fromList(m_pWorldHandle->selectionSetHandle()->occurencesList());
    QList<GLC_StructOccurence *> subject;
    QSet<GLC_StructOccurence *>::ConstIterator iOcc= selectedOccSet.begin();
    while (iOcc != selectedOccSet.constEnd())
    {
        QList<GLC_StructOccurence*> ancestorList= (*iOcc)->ancestorList();
        const int count= ancestorList.count();
        bool addOcc= true;
        for (int i= 0; i < count; ++i)
        {
            if (selectedOccSet.contains(ancestorList.at(i)))
            {
                addOcc= false;
                break;
            }
        }
        if (addOcc) subject.append(*iOcc);

        ++iOcc;
    }

    return subject;
}

GLC_StructOccurence* GLC_World::takeRootOccurrence()
{
	GLC_StructOccurence* pSubject= m_pRoot;
	pSubject->makeOrphan();

	m_pRoot= new GLC_StructOccurence();
	m_pRoot->setWorldHandle(m_pWorldHandle);

	return pSubject;
}

void GLC_World::replaceRootOccurrence(GLC_StructOccurence* pOcc)
{
	Q_ASSERT(pOcc->isOrphan());
	delete m_pRoot;
	m_pRoot= pOcc;
	m_pRoot->setWorldHandle(m_pWorldHandle);
}

void GLC_World::mergeWithAnotherWorld(GLC_World& anotherWorld)
{
	GLC_StructOccurence* pAnotherRoot= anotherWorld.rootOccurence();
	if (pAnotherRoot->childCount() > 0)
	{
		QList<GLC_StructOccurence*> childs= pAnotherRoot->children();
		const int size= childs.size();
		for (int i= 0; i < size; ++i)
		{
			m_pRoot->addChild(childs.at(i)->clone(m_pWorldHandle, false));
		}
		m_pRoot->updateChildrenAbsoluteMatrix();
	}
	else
	{
		m_pRoot->addChild(anotherWorld.rootOccurence()->clone(m_pWorldHandle, false));
	}
}

GLC_World& GLC_World::operator=(const GLC_World& world)
{
    if ((this != &world) && (this->m_pWorldHandle != world.m_pWorldHandle))
	{
        // Decrement the number of world
        m_pWorldHandle->decrement();
        if (m_pWorldHandle->isOrphan())
        {
            delete m_pRoot;
            delete m_pWorldHandle;
        }
        m_pRoot= world.m_pRoot;
        m_pWorldHandle= world.m_pWorldHandle;
        m_pWorldHandle->increment();
    }
    return *this;
}

bool GLC_World::operator ==(const GLC_World &other) const
{
    bool subject= this == &other;
    subject= (subject || (this->m_pWorldHandle == other.m_pWorldHandle));

    return subject;
}
